/*********************************************
********** Reading and writing to a **********
********* MLX90614 sensor via SMBus **********
************** by Matej Nogic ****************
*** Zagreb University of Applied Sciences ****
**********************************************
**********************************************

Program is controlled over UART, baud rate: 19200bps
Write "Read" to read all the register's values
Write "Write" to write settings listed from line 178

If you want to switch from PWM mode back to SMBus mode
you need to pull SCL pin low for a time longer than 1.024ms.
This will allow you to communicate with a sensor over SMBus
but the PWMCTRL register value will NOT change. So if you want to
stay in a SMBus mode after POR you have to change the PWMCTRL
register value manually.
*/


//Addresses
const mlx90614_address = 0x5A;
const amb_temp = 0x06;
const obj_temp = 0x07;
const Tmax = 0x20;
const Tmin = 0x21;
const PWMCTRL = 0x22;
const amb_range = 0x23;
const Conreg1 = 0x25;
const SMBus_Address = 0x2E;

unsigned char txt[30],uart_rd[30],counter = 0 ;

int write_mlx90614(char address, long value){
    int a,cnt=0;
    char lsb=0,msb=0;

    if (address >= 0x20){   //Only write in EEPROM
        if (address <= 0x21) value = value * 100 + 27315;  //Tmax and Tmin are given in degrees celsius

        switch(address){
           case Tmax: UART1_Write_Text("Writing Tmax...\r\n"); break;
           case Tmin: UART1_Write_Text("Writing Tmin...\r\n"); break;
           case amb_range: UART1_Write_Text("Writing Ta range...\r\n"); break;
           case PWMCTRL: UART1_Write_Text("Writing PWMCTRL...\r\n"); break;
           case Conreg1: UART1_Write_Text("Writing Config Register 1...\r\n"); break;
           case SMBus_Address: UART1_Write_Text("Writing SMBus Address...\r\n"); break;
        }

        while(cnt < 2){          //2 rounds: erase and write new data
            if (cnt == 1){
                delay_ms(5);      //Time needed for cell to be erased (Melexis Application Note)
                lsb = value & 0xFF;
                msb = value >> 8;
            }
            // Finding CRC
            for(a=0; a<256; a++){
                I2C1_Start();
                I2C1_Wr(mlx90614_address << 1);
                I2C1_Wr(address);
                I2C1_Wr(lsb);
                I2C1_Wr(msb);

                if (I2C1_Wr(a) == 0){
                    UART1_Write_Text("CRC found!\r\n");
                    UART1_Write_Text("CRC: ");
                    InttoStr(a,txt);
                    UART1_Write_Text(Ltrim(txt));
                    UART1_Write_Text("\r\n");
                }
            }
            I2C1_Stop();
            cnt++;
        }
        UART1_Write_Text("Data written\r\n\r\n");
    }
    else UART1_Write_Text("Can't write to RAM\r\n\r\n");

    return 0;
}

float read_mlx90614(char address){
    unsigned int value;
    float value_celsius;

    I2C1_Start();
    I2C1_Wr(mlx90614_address << 1);
    I2C1_Wr(address);
    I2C1_Repeated_Start();
    I2C1_Wr((mlx90614_address << 1)+1);
    value = I2C1_Rd(0);
    if (address != SMBus_Address){    //SMBus address is 8bit
       value = (I2C1_Rd(0) << 8) | value;
    }
    I2C1_Stop();

    switch(address){
         case obj_temp: UART1_Write_Text("To = "); break;
         case amb_temp: UART1_Write_Text("Ta = "); break;
         case Tmax: UART1_Write_Text("Tmax = "); break;
         case Tmin: UART1_Write_Text("Tmin = "); break;
         case amb_range: UART1_Write_Text("Ta range = "); break;
         case PWMCTRL: UART1_Write_Text("PWMCTRL = "); break;
         case Conreg1: UART1_Write_Text("Config Register 1 = "); break;
         case SMBus_Address: UART1_Write_Text("SMBus Address = "); break;
    }

    if (address <= 0x21){
        if (address <= 0x07){
             value_celsius = value * 0.02 - 273.15;
        }
        else value_celsius = value / 100.0 - 273.15;
        FloatToStr(value_celsius, txt);

        UART1_Write_Text(Ltrim(txt));
        UART1_Write_Text(" C (Register value: ");

        WordToStr(value, txt);
        UART1_Write_Text(Ltrim(txt));
        UART1_Write_Text(") \r\n");
    }
    else if (address == amb_range){
        value_celsius = (((value & 0xFF) * 64.0) / 100.0) - 38.2;
        FloatToStr(value_celsius, txt);
        UART1_Write_Text(Ltrim(txt));

        UART1_Write_Text(" - ");

        value_celsius = (((value >> 8) * 64.0) / 100.0) - 38.2;
        FloatToStr(value_celsius, txt);
        UART1_Write_Text(Ltrim(txt));
        UART1_Write_Text(" C (Register value: ");

        WordToStr(value, txt);
        UART1_Write_Text(Ltrim(txt));
        UART1_Write_Text(") \r\n");
    }
    else{
        WordToHex(value, txt);
        UART1_Write_Text("0x");
        UART1_Write_Text(txt);
        UART1_Write_Text("\r\n");
    }

    return 0;
}

void main() {
    I2C1_Init(100000);
    Delay_ms(500);

    UART1_Init(19200);
    Delay_ms(500);

    UART1_Write_Text("Ready\r\n");

    while (1){
       if(UART1_Data_Ready()){
            switch(UART1_Read()){
                 case '\n':
                      if(strcmp(uart_rd, "Read") == 0){
                           UART1_Write_Text("Reading...\r\n\r\n");
                           UART1_Write_Text("RAM:\r\n") ;
                           read_mlx90614(obj_temp);
                           read_mlx90614(amb_temp);
                           UART1_Write_Text("\r\nEEPROM:\r\n") ;
                           read_mlx90614(Tmax);
                           read_mlx90614(Tmin);
                           read_mlx90614(amb_range);
                           read_mlx90614(PWMCTRL);
                           read_mlx90614(Conreg1);
                           read_mlx90614(SMBus_Address);

                           UART1_Write_Text("\r\n\r\n") ;
                      }
                      else if(strcmp(uart_rd, "Write") == 0){
                        /* PUT HERE WHAT YOU WANT TO BE WRITTEN
                      Tmin and Tmax should be written in degrees Celsius,
                       others registers should be written in raw format.
                       Register's addresses are at the beginning of the code */
                       
                           write_mlx90614(Tmax,100);  // Set Tmax to 100C
                           write_mlx90614(Tmin,0);    // Set Tmin to 0C
                           write_mlx90614(PWMCTRL,513);  //Set PWMCTRL register to a value 513 decimal (PWM OFF)
                      }
                      else UART1_Write_Text("Unknown command. Use Read and Write.\r\n") ;

                      counter = 0 ;
                      *uart_rd = 0 ;
                      break ;

                  case '\r': break ;

                  default:
                       uart_rd[counter] = UART1_Read() ;
                       counter++ ;
                       uart_rd[counter] = 0 ;
            }
       }
    }
}